class Feeds(object):
    """
    The Amazon MWS Feeds API section of the Amazon Marketplace
    Web Service (Amazon MWS) API lets you upload inventory and
    order data to Amazon.

    .. code-block:: python

        response = client.feeds.submit_feed(
            feed_content_xml,
            FeedType='_POST_INVENTORY_AVAILABILITY_DATA_'
        )
        result = client.feeds.get_feed_submission_result(
            response.FeedSubmissionInfo.FeedSubmissionId
        )

    """
    VERSION = '2009-01-01'
    URI = '/Feeds/' + VERSION

    def __init__(self, client):
        self.client = client

    def submit_feed(self, FeedContent, FeedType, ContentType=None, **kwargs):
        """
        Uploads a feed for processing by Amazon MWS.

        :param FeedContent: String or base string content
        :param FeedType: Type of feed
                         (see `MWS docs <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_FeedType.html>`_)
        :param ContentType: Set the content type of feed.
                            (See `Feedtypes <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_SubmitFeed.html>`_).
                            If not set, this will be guessed based on
                            FeedType

        `Learn more <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_SubmitFeed.html>`_
        """    # noqa: E501
        kwargs['FeedType'] = FeedType
        if ContentType is None:
            if '_FLAT_FILE_' in FeedType:
                ContentType = \
                    'text/tab-separated-values; charset=iso-8859-1'
                FeedContent = FeedContent.encode('iso-8859-1')
            else:
                ContentType = 'text/xml'
        return self.client.post(
            'SubmitFeed', self.URI, kwargs, self.VERSION,
            FeedContent, ContentType
        )

    def get_feed_submission_list(self, **kwargs):
        """
        Returns a list of all feed submissions submitted in the previous 90 days.

        `Learn more <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_GetFeedSubmissionList.html>`_
        """    # noqa: E501
        return self.client.get(
            'GetFeedSubmissionList', self.URI, kwargs, self.VERSION
        )

    def get_feed_submission_list_by_next_token(self, NextToken):
        """
        Returns a list of feed submissions using the NextToken parameter.

        `Learn more <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_GetFeedSubmissionListByNextToken.html>`_
        """    # noqa: E501
        return self.client.get(
            'GetFeedSubmissionListByNextToken', self.URI,
            {'NextToken': NextToken}, self.VERSION
        )

    def get_feed_submission_count(self, **kwargs):
        """
        Returns a count of the feeds submitted in the previous 90 days.

        `Learn more <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_GetFeedSubmissionCount.html>`_
        """    # noqa: E501
        return self.client.get(
            'GetFeedSubmissionCount', self.URI, kwargs, self.VERSION
        )

    def cancel_feed_submissions(self, **kwargs):
        """
        Cancels one or more feed submissions and returns a count of the
        feed submissions that were canceled.

        `Learn more <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_CancelFeedSubmissions.html>`_
        """    # noqa: E501
        return self.client.get(
            'CancelFeedSubmissions', self.URI, kwargs, self.VERSION
        )

    def get_feed_submission_result(self, FeedSubmissionId):
        """
        Returns the feed processing report and the Content-MD5 header.

        `Learn more <https://docs.developer.amazonservices.com/en_US/feeds/Feeds_GetFeedSubmissionResult.html>`_
        """    # noqa: E501
        return self.client.get(
            'GetFeedSubmissionResult', self.URI,
            {'FeedSubmissionId': FeedSubmissionId},
            self.VERSION
        )



import base64
from collections import namedtuple
import hashlib

from builtins import str
import six

from .exceptions import MWSException

if six.PY2:
    import unicodecsv as csv
    from io import BytesIO
else:
    import csv


Marketplace = namedtuple(
    'Marketplace', ['code', 'currency', 'id', 'endpoint', 'name']
)

#: A list of `Marketplaces
#: <http://docs.developer.amazonservices.com/en_US/dev_guide/DG_Endpoints.html>`_
MARKETPLACES = [

    # North america region
    Marketplace(
        'BR', 'BRL', 'A2Q3Y263D00KWC',
        'https://mws.amazonservices.com', 'Amazon.com.br'),
    Marketplace(
        'CA', 'CAD', 'A2EUQ1WTGCTBG2',
        'https://mws.amazonservices.ca', 'Amazon.ca'),
    Marketplace(
        'MX', 'MXN', 'A1AM78C64UM0Y8',
        'https://mws.amazonservices.com.mx', 'Amazon.com.mx'),
    Marketplace(
        'US', 'USD', 'ATVPDKIKX0DER',
        'https://mws.amazonservices.com', 'Amazon.com'),

    # Europe region
    Marketplace(
        'AE', 'AED', 'A2VIGQ35RCS4UG',
        'https://mws.amazonservices.ae', 'United Arab Emirates (U.A.E.)'),
    Marketplace(
        'DE', 'EUR', 'A1PA6795UKMFR9',
        'https://mws-eu.amazonservices.com', 'Amazon.de'),
    Marketplace(
        'EG', 'EGP', 'ARBP9OOSHTCHU',
        'https://mws-eu.amazonservices.com', 'Egypt'),
    Marketplace(
        'ES', 'EUR', 'A1RKKUPIHCS9HS',
        'https://mws-eu.amazonservices.com', 'Amazon.es'),
    Marketplace(
        'FR', 'EUR', 'A13V1IB3VIYZZH',
        'https://mws-eu.amazonservices.com', 'Amazon.fr'),
    Marketplace(
        'GB', 'GBP', 'A1F83G8C2ARO7P',
        'https://mws-eu.amazonservices.com', 'Amazon.co.uk'),
    Marketplace(
        'IN', 'INR', 'A21TJRUUN4KGV',
        'https://mws.amazonservices.in', 'Amazon.in'),
    Marketplace(
        'IT', 'EUR', 'APJ6JRA9NG5V4',
        'https://mws-eu.amazonservices.com', 'Amazon.it'),
    Marketplace(
        'NL', 'EUR', 'A1805IZSGTT6HS',
        'https://mws-eu.amazonservices.com', 'Amazon.nl'),
    Marketplace(
        'SA', 'SAR', 'A17E79C6D8DWNP',
        'https://mws-eu.amazonservices.com', 'Saudi Arabia'),
    Marketplace(
        'TR', 'EUR', 'A33AVAJ2PDY3EV',
        'https://mws-eu.amazonservices.com', 'Turkey'),

    # Far East region
    Marketplace(
        'SG', 'SGD', 'A19VAU5U5O7RUS',
        'https://mws-fe.amazonservices.com', 'Amazon.sg'),
    Marketplace(
        'AU', 'AUD', 'A39IBJ37TRP1C6',
        'https://mws.amazonservices.com.au', 'Amazon.com.au'),
    Marketplace(
        'JP', 'JPY', 'A1VC38T7YXB528',
        'https://mws.amazonservices.jp', 'Amazon.jp'),

]


def get_marketplace(id_or_code):
    """
    Given a code, get the marketplace
    """
    id_or_code = id_or_code.upper()

    for marketplace in MARKETPLACES:
        if marketplace.code == id_or_code or \
                marketplace.id == id_or_code:
            return marketplace

    raise MWSException(
        '{} is not a valid marketplace id or code'.format(id_or_code)
    )


def flatten_list(kwargs, key, separator):
    """
    Convert a list into URL parameters the way amazon like it.

    Example::

        flatten_list(
            {'ReportTypeList': ['_A_', '_B_']},
            'ReportTypeList',
            'Type'
        )

    Becomes::

        ReportTypeList.Type.1=_A_&ReportTypeList.Type.2=_B_
    """
    vals = kwargs.pop(key, [])
    for index, value in enumerate(vals, 1):
        subkey = '{}.{}.{}'.format(key, separator, index)
        kwargs[subkey] = value
        if isinstance(value, dict):
            flatten_dict(kwargs, subkey)


def flatten_dict(kwargs, key):
    """
    Convert a dict into URL parameters the way amazon like it.

    Example::

        flatten_dict(
            {'DestinationAddress': {'Name': 'John', 'Country': 'US'}},
            'DestinationAddress',
        )

    Becomes::

        DestinationAddress.Name=John&DestinationAddress.Country=US
    """
    vals = kwargs.pop(key, {})
    for subkey, value in vals.items():
        kwargs['{}.{}'.format(key, subkey)] = value


def parse_tsv(text):
    """
    Python 2 and 3 compatible TSV parser that returns a list of
    dictionary objects.
    """
    if six.PY2:
        text = BytesIO(text.encode('utf-8'))
    else:
        text = text.splitlines()
    return list(
        csv.DictReader(text, dialect='excel-tab')
    )


def get_md5_hash(string):
    hasher = hashlib.md5()
    if isinstance(string, str):
        hasher.update(string.encode('utf-8'))
    else:
        hasher.update(string)
    return base64.b64encode(
        hasher.digest()
    ).decode('utf-8')



"""
All errors and exceptions
"""


class MWSException(Exception):
    """
    Parent class of all exceptions this package handles
    """
    pass


class MWSError(MWSException):
    """
    Parent class of all Amazon returned Errors.
    """
    pass


class AccessDenied(MWSError):
    "Access was denied"


class SignatureDoesNotMatch(MWSError):
    """The signature used does not match the server's
    calculated signature value."""
    pass


class QuotaExceeded(MWSError):
    """The total number of requests in an hour was exceeded."""
    pass


class RequestThrottled(MWSError):
    """The frequency of requests was greater than allowed.."""
    pass




def test_submit_feed(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'POST',
        mws_client.marketplace.endpoint + '/Feeds/2009-01-01',
        status_code=200,
        text=example_response('feeds/submit_feed.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.feeds.submit_feed(
        '<mock></mock>',
        '_POST_ORDER_ACKNOWLEDGEMENT_DATA_'
    )
    assert response.FeedSubmissionInfo.FeedSubmissionId
    assert response.FeedSubmissionInfo.FeedProcessingStatus == '_SUBMITTED_'

    response = mws_client.feeds.submit_feed(
        b'<mock></mock>',
        '_POST_ORDER_ACKNOWLEDGEMENT_DATA_'
    )
    assert response.FeedSubmissionInfo.FeedSubmissionId
    assert response.FeedSubmissionInfo.FeedProcessingStatus == '_SUBMITTED_'


def test_feed_result(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Feeds/2009-01-01',
        status_code=200,
        text=example_response('feeds/result-failed.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.feeds.get_feed_submission_result('4901340')
    assert response.MessageType == 'ProcessingReport'


def test_feed_submission_count(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Feeds/2009-01-01',
        status_code=200,
        text=example_response('feeds/submission-count.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.feeds.get_feed_submission_count()
    assert response.Count == 463


def test_feed_submission_cancel(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Feeds/2009-01-01',
        status_code=200,
        text=example_response('feeds/submission-cancel.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.feeds.cancel_feed_submissions()
    assert response.Count == 1


def test_feed_submission_list(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Feeds/2009-01-01',
        status_code=200,
        text=example_response('feeds/submission-list.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.feeds.get_feed_submission_list()
    assert response.NextToken
    assert response.FeedSubmissionInfo.FeedType == '_POST_PRODUCT_DATA_'


def test_feed_submission_list_next(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Feeds/2009-01-01',
        status_code=200,
        text=example_response('feeds/submission-list-token.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.feeds.get_feed_submission_list_by_next_token(
        'next-token'
    )
    assert response.NextToken == 'none'
    assert response.FeedSubmissionInfo.FeedType == '_POST_PRODUCT_DATA_'



from io import open
import os

import pytest
import requests_mock

from pymws import MWS


@pytest.fixture()
def mws_client():
    return MWS(
        'US',
        access_key_id='ACESSKEY',
        secret_key='SECRET',
        merchant_id='MERCHANT_ID',
        auth_token='amzn.mws.big-fat-token',
    )


@pytest.fixture()
def mock_adapter(mws_client):
    adapter = requests_mock.Adapter()
    mws_client.session.mount(mws_client.marketplace.endpoint, adapter)
    return adapter


@pytest.fixture()
def example_response():
    def get_example_response(path):
        directory = os.path.dirname(
            os.path.abspath(__file__)
        )
        with open(
                os.path.join(directory, 'responses', path),
                encoding='utf-8') as f:
            return f.read()
    return get_example_response



"""Top-level package for Python MWS Client."""

__author__ = """Fulfil.IO Inc."""
__email__ = 'help@fulfil.io'
__version__ = '0.1.5'

from .pymws import MWS              # noqa
from .exceptions import MWSError    # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from datetime import datetime

from click.testing import CliRunner
import pytest

from pymws import MWS, cli
from pymws.exceptions import AccessDenied
from pymws.utils import flatten_list, flatten_dict


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'pymws.cli.main' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output


def test_query_string_builder():
    params = {
        'Name': 'ST',
        'City': 'Los Ángeles',
        'Timestamp': datetime(2020, 1, 1, 10, 30, 30)
    }
    client = MWS('US')
    assert client.build_query_string(params) == \
        'City=Los%20%C3%81ngeles&Name=ST&Timestamp=2020-01-01T10%3A30%3A30'


def test_list_flattening():
    kwargs = {'ReportTypeList': ['_A_', '_B_']}
    flatten_list(
        kwargs,
        'ReportTypeList',
        'Type'
    )
    assert kwargs['ReportTypeList.Type.1'] == '_A_'
    assert kwargs['ReportTypeList.Type.2'] == '_B_'

    kwargs = {}
    flatten_list(
        kwargs,
        'InexistentKey',
        'Type'
    )
    assert not kwargs


def test_dict_flattening():
    kwargs = {
        'DestinationAddress': {
            'Name': 'John Doe',
            'Line1': 'Random street',
            'StateOrProvinceCode': 'CA',
            'CountryCode': 'US',
        }
    }
    flatten_dict(
        kwargs,
        'DestinationAddress',
    )
    assert kwargs['DestinationAddress.Name'] == 'John Doe'
    assert kwargs['DestinationAddress.CountryCode'] == 'US'

    kwargs = {}
    flatten_dict(
        kwargs,
        'DestinationAddress',
    )
    assert not kwargs


def test_get_service_status(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Orders/2013-09-01',
        status_code=401,
        text=example_response('401.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    with pytest.raises(AccessDenied):
        mws_client.orders.get_service_status()



from ..utils import flatten_list, flatten_dict


class OutboundShipment(object):
    """
    Implements outbound shipment client for Amazon MWS

    .. code-block:: python

        client.fulfillment_outbound_shipment.create_fulfillment_order(
            MarketplaceId='A2Q3Y263D00KWC',
            SellerFulfillmentOrderId='SO3421',
            FulfillmentAction='Ship',
            DisplayableOrderId='SO3421',
            DisplayableOrderDateTime=datetime.utcnow(),
            DisplayableOrderComment='Some comment',
            ShippingSpeedCategory='Standard',
            DestinationAddress={
                'Name': 'John Doe',
                'Line1': 'Random street',
                'StateOrProvinceCode': 'CA',
                'CountryCode': 'US',
            },
            Items=[{
                'SellerSKU': 'SKU-1',
                'SellerFulfillmentOrderItemId': 'SO3421-1',
                'Quantity': 1,
            }],
        )

    for other attributes refer to the Amazon MWS documentation.
    """
    VERSION = "2010-10-01"
    URI = "/FulfillmentOutboundShipment/" + VERSION

    def __init__(self, client):
        self.client = client

    def create_fulfillment_order(self, **kwargs):
        """
        Requests that Amazon ship items from the seller's inventory in Amazon's
        fulfillment network to a destination address.

        `Learn more <http://docs.developer.amazonservices.com/en_US/fba_outbound/FBAOutbound_CreateFulfillmentOrder.html>`__
        """  # noqa: E501
        flatten_dict(kwargs, 'DestinationAddress')
        for item in kwargs.get('Items'):
            flatten_dict(item, 'PerUnitDeclaredValue')
        flatten_list(kwargs, 'Items', 'member')
        flatten_list(kwargs, 'NotificationEmailList', 'member')
        return self.client.post(
            'CreateFulfillmentOrder', self.URI, kwargs, self.VERSION
        )

    def get_fulfillment_order(self, SellerFulfillmentOrderId):
        """
        Returns a fulfillment order based on a specified SellerFulfillmentOrderId.

        `Learn more <Learn more: http://docs.developer.amazonservices.com/en_US/fba_outbound/FBAOutbound_GetFulfillmentOrder.html>`__
        """  # noqa: E501
        return self.client.get(
            'GetFulfillmentOrder', self.URI,
            {'SellerFulfillmentOrderId': SellerFulfillmentOrderId},
            self.VERSION
        )



from datetime import datetime


def test_request_report(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'POST',
        mws_client.marketplace.endpoint + '/Reports/2009-01-01',
        status_code=200,
        text=example_response('reports/request_report.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.reports.request_report(
        Marketplace='ATVPDKIKX0DER',
        ReportType='_GET_MERCHANT_LISTINGS_DATA_',
        StartDate=datetime(2009, 1, 3, 18, 12, 21)
    )
    assert response.ReportRequestInfo.ReportProcessingStatus == '_SUBMITTED_'


def test_report_request_list(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Reports/2009-01-01',
        status_code=200,
        text=example_response('reports/get_report_request_list.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.reports.get_report_request_list(
        Marketplace='ATVPDKIKX0DER',
        ReportTypeList=[
            '_GET_ORDERS_DATA_',
            '_GET_MERCHANT_LISTINGS_DATA_',
        ],
        ReportProcessingStatusList=[
            '_DONE_',
        ],
        ReportRequestIdList=['2291326454']
    )

    assert 'ReportTypeList.Type.1=_GET_ORDERS_DATA_' in \
        mock_adapter.request_history[0].url

    assert response.ReportRequestInfo.ReportProcessingStatus == '_DONE_'
    assert response.NextToken


def test_report_request_list_by_next_token(
        mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Reports/2009-01-01',
        status_code=200,
        text=example_response(
            'reports/get_report_request_list_by_next_token.xml'
        ),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.reports.get_report_request_list_by_next_token(
        NextToken='2YgYW55IPQhcm5hbCBwbGVhc3VyZS4='
    )
    assert response.ReportRequestInfo.ReportProcessingStatus == '_SUBMITTED_'
    assert response.NextToken


def test_get_report_tsv(
        mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Reports/2009-01-01',
        status_code=200,
        text=example_response(
            'reports/get_report_tab_separated.tsv'
        ),
        headers={'Content-Type': 'text/plain;charset=Cp1252'}
    )
    response = mws_client.reports.get_report(123456789)
    assert len(response) == 3
    assert response[0]['listing-id'] == '0305XXNBYUQ'



from .utils import flatten_list


class Reports(object):
    """
    Implements a report client for Amazon MWS.

    The report API works differently from other APIs and
    it is important to have a good understanding of
    how this API works from `MWS reports overview
    <http://docs.developer.amazonservices.com/en_US/reports/Reports_Overview.html>`__
    """    # noqa: E501
    VERSION = '2009-01-01'
    URI = '/Reports/' + VERSION

    def __init__(self, client):
        self.client = client

    def request_report(self, **kwargs):
        """
        Creates a report request and submits the request to Amazon MWS.

        Amazon MWS processes the report request and when the report is
        completed, sets the status of the report request to _DONE_.
        Reports are retained for 90 days.

        `Learn more <Read more: http://docs.developer.amazonservices.com/en_US/reports/Reports_RequestReport.html>`__
        """   # noqa: E501
        return self.client.post(
            'RequestReport', self.URI, kwargs, self.VERSION
        )

    def get_report_request_list(self, **kwargs):
        """
        Returns a list of report requests that you can use to get the
        ReportRequestId for a report.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReportRequestList.html>`__
        """   # noqa: E501
        flatten_list(kwargs, 'ReportTypeList', 'Type')
        flatten_list(kwargs, 'ReportRequestIdList', 'Id')
        flatten_list(kwargs, 'ReportProcessingStatusList', 'Status')
        return self.client.get(
            'GetReportRequestList', self.URI, kwargs, self.VERSION
        )

    def get_report_request_list_by_next_token(self, NextToken):
        """
        Returns a list of report requests using the NextToken,
        which was supplied by a previous request to either
        GetReportRequestListByNextToken or GetReportRequestList,
        where the value of HasNext was true in that previous request.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReportRequestListByNextToken.html>`__
        """   # noqa: E501
        return self.client.get(
            'GetReportRequestListByNextToken', self.URI,
            {'NextToken': NextToken}, self.VERSION
        )

    def get_report_request_count(self, **kwargs):
        """
        Returns a count of report requests that have been submitted
        to Amazon MWS for processing.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReportRequestCount.html>`__
        """   # noqa: E501
        return self.client.get(
            'GetReportRequestCount', self.URI, kwargs, self.VERSION
        )

    def cancel_report_request(self, **kwargs):
        """
        Cancels one or more report requests.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_CancelReportRequests.html>`__
        """  # noqa: E501
        return self.client.post(
            'CancelReportRequests', self.URI, kwargs, self.VERSION
        )

    def get_report_list(self, **kwargs):
        """
        Returns a list of reports that were created in the previous 90 days.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReportList.html>`__
        """   # noqa: E501
        flatten_list(kwargs, 'ReportTypeList', 'Type')
        flatten_list(kwargs, 'ReportRequestIdList', 'Id')
        return self.client.get(
            'GetReportList', self.URI, kwargs, self.VERSION
        )

    def get_report_list_by_next_token(self, NextToken):
        """
        Returns a list of reports using the NextToken, which was supplied
        by a previous request to either GetReportListByNextToken or
        GetReportList, where the value of HasNext was true in the
        previous call.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReportListByNextToken.html>`__
        """   # noqa: E501
        return self.client.get(
            'GetReportListByNextToken', self.URI,
            {'NextToken': NextToken}, self.VERSION
        )

    def get_report_count(self, **kwargs):
        """
        Returns a count of the reports, created in the previous 90 days,
        with a status of _DONE_ and that are available for download.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReportCount.html>`__
        """   # noqa: E501
        return self.client.get(
            'GetReportCount', self.URI, kwargs, self.VERSION
        )

    def get_report(self, ReportId):
        """
        Returns the contents of a report and the Content-MD5 header for the
        returned report body.

        `Learn more <http://docs.developer.amazonservices.com/en_US/reports/Reports_GetReport.html>`__
        """   # noqa: E501
        return self.client.get(
            'GetReport', self.URI,
            {'ReportId': ReportId}, self.VERSION
        )



"""Main module."""
from __future__ import unicode_literals

import base64
import hashlib
import hmac
from builtins import str as text
from datetime import date, datetime

import requests
from lxml import objectify

from .exceptions import MWSError, AccessDenied, QuotaExceeded, RequestThrottled
from .feeds import Feeds
from .orders import Orders
from .products import Products
from .reports import Reports
from .fulfillment.outbound_shipment import OutboundShipment
from .utils import get_marketplace, parse_tsv, get_md5_hash

try:
    from urllib.parse import urlparse, quote
except ImportError:
    # py2
    from urlparse import urlparse
    from urllib import quote


MWS_SAFE = '-_.~'.encode('utf-8')


class MWS(object):
    """
    Primary client class that acts as a gateway to all of the
    functionality offered by this package.

    Example usage::

        client = MWS(
            marketplace="US", merchant_id="1234",
            access_key_id="key", secret_key="secret",
            auth_token="token"
        )
        client.reports.get_reports_list()

    :param marketplace: marketplace to connect to.
    :param merchant_id: Amazon merchant it.
    :param access_key_id: Access key of your app
    :param secret_key: Secret key of your app
    :param auth_token: Token obtained by the merchant after
                       installing your app.
    """

    def __init__(
                self,
                marketplace, merchant_id=None,
                access_key_id=None, secret_key=None,
                auth_token=None):
        self.marketplace = get_marketplace(marketplace)
        self.merchant_id = merchant_id
        self.access_key_id = access_key_id
        self.secret_key = secret_key
        self.auth_token = auth_token
        self.session = requests.Session()
        self.user_agent = 'pymws/0.1 (Language=Python)'

    @property
    def products(self):
        """
        Fetch the products API client.
        Returns an instance of :class:`pymws.products.Products`
        """
        return Products(self)

    @property
    def orders(self):
        """
        Fetch the order API client
        Returns an instance of :class:`pymws.orders.Orders`
        """
        return Orders(self)

    @property
    def reports(self):
        """
        Fetch the reports API client
        Returns an instance of :class:`pymws.reports.Reports`
        """
        return Reports(self)

    @property
    def feeds(self):
        """
        Fetch the feeds API client
        Returns an instance of :class:`pymws.feeds.Feeds`
        """
        return Feeds(self)

    @property
    def fulfillment_outbound_shipment(self):
        """
        Fetch the Fulfillment outbound shipment API client
        Returns an instance of
        :class:`pymws.fulfillment.outbound_shipment.OutboundShipment`
        """
        return OutboundShipment(self)

    def get(self, action, uri, req_params, version):
        return self._request(
            'GET',
            action, uri, req_params, version
        )

    def post(self, action, uri, req_params, version,
             body=None, content_type=None):
        return self._request(
            'POST',
            action, uri, req_params, version, body, content_type
        )

    def _request(self, http_verb, action, uri, req_params, version,
                 body=None, content_type=None):
        """
        Build a request, parse the response and handle errors
        """
        if body is not None:
            req_params['ContentMD5Value'] = get_md5_hash(body)

        query_string = self.get_query_string(action, req_params, version)
        signature = self.get_signature(http_verb, uri, query_string)
        url = "{endpoint}{uri}".format(
            endpoint=self.marketplace.endpoint,
            uri=uri,
        )
        headers = {
            'User-Agent': self.user_agent,
        }

        if content_type:
            headers['Content-Type'] = content_type

        response = self.session.request(
            http_verb,
            url,
            data=body,
            params="{query_string}&Signature={signature}".format(
                query_string=query_string,
                signature=quote(signature)
            ),
            headers=headers
        )

        if response.status_code == 401:
            raise AccessDenied(response.text)
        elif response.status_code == 503:
            error_code = objectify.fromstring(response.text).Error.Code
            if error_code == 'RequestThrottled':
                raise RequestThrottled(response.text)
            elif error_code == 'QuotaExceeded':
                raise QuotaExceeded(response.text)
        elif response.status_code != 200:
            raise MWSError(response.text)

        if response.headers['content-type'].startswith('text/xml'):
            xml = objectify.fromstring(response.content)
            result_el = '{}Result'.format(action)
            if hasattr(xml, result_el):
                return getattr(xml, result_el)
            return xml
        elif response.headers['content-type'].startswith('text/plain'):
            return parse_tsv(response.content.decode(encoding="iso-8859-1"))
        else:
            return response.text

    def get_query_string(self, action, req_params, version):
        params = {
            'AWSAccessKeyId': self.access_key_id,
            'Action': action,
            'MWSAuthToken': self.auth_token,
            'SellerId': self.merchant_id,
            'SignatureMethod': 'HmacSHA256',
            'SignatureVersion': '2',
            'Timestamp': datetime.utcnow().isoformat(),
            'Version': version,
        }
        params.update(req_params)
        query_string = self.build_query_string(params)
        return query_string

    def get_signature(self, http_verb, uri, query_string):
        to_sign = [
            http_verb,
            urlparse(self.marketplace.endpoint).netloc,
            uri,
            query_string
        ]
        return base64.b64encode(
            hmac.new(
                self.secret_key.encode('utf-8'),
                '\n'.join(to_sign).encode('utf-8'),
                hashlib.sha256,
            ).digest()
        ).decode('utf-8')

    def build_query_string(self, params):
        """
        create the query string to be signed

        * Sort the UTF-8 query string components by parameter name
        * URL encode the parameter name and values

        Because of the following issues, this method cannot directly
        call urlencode because urlencode does not support quoting the
        way amazon wants, it always does quote_plus.
        """
        query = []
        for key in sorted(params.keys()):
            value = params[key]

            if isinstance(value, (date, datetime)):
                value = value.isoformat()

            if not isinstance(value, text):
                value = str(value)

            query.append(
                '{}={}'.format(
                    quote(key, safe=MWS_SAFE),
                    quote(value.encode('utf-8'), safe=MWS_SAFE)
                )
            )
        return '&'.join(query)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from datetime import datetime


def test_create_fulfillment_order(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'POST', (
            mws_client.marketplace.endpoint +
            '/FulfillmentOutboundShipment/2010-10-01'
        ),
        status_code=200,
        text=example_response(
            'outbound_shipment/create_fulfillment_order.xml'
        ),
        headers={'Content-Type': 'text/xml'}
    )
    fulfillment_outbound_shipment = mws_client.fulfillment_outbound_shipment
    response = fulfillment_outbound_shipment.create_fulfillment_order(
        MarketplaceId='A2Q3Y263D00KWC',
        SellerFulfillmentOrderId='SO3421',
        FulfillmentAction='Ship',
        DisplayableOrderId='SO3421',
        DisplayableOrderDateTime=datetime.utcnow(),
        DisplayableOrderComment='Some comment',
        ShippingSpeedCategory='Standard',
        DestinationAddress={
            'Name': 'René Magritte',
            'Line1': 'Random street',
            'StateOrProvinceCode': 'CA',
            'CountryCode': 'US',
        },
        Items=[{
            'SellerSKU': 'SKU-1',
            'SellerFulfillmentOrderItemId': 'SO3421-1',
            'Quantity': 1,
        }, {
            'SellerSKU': 'SKU-2',
            'SellerFulfillmentOrderItemId': 'SO3421-2',
            'Quantity': 10.0,
            'PerUnitDeclaredValue': {
                'CurrencyCode': 'USD',
                'Value': "32",
            }
        }],
    )
    assert response.ResponseMetadata.RequestId
    url = mock_adapter.request_history[0].url
    assert 'DestinationAddress.Name=Ren%C3%A9%20Magritte' in url
    assert 'Items.member.1.SellerSKU=SKU-1' in url
    assert 'Items.member.1.Quantity=1' in url
    assert 'Items.member.2.SellerSKU=SKU-2' in url
    assert 'Items.member.2.Quantity=10.0' in url
    assert 'Items.member.2.PerUnitDeclaredValue.CurrencyCode=USD' in url
    assert 'Items.member.2.PerUnitDeclaredValue.Value=32' in url


def test_get_fulfillment_order(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET', (
            mws_client.marketplace.endpoint +
            '/FulfillmentOutboundShipment/2010-10-01'
        ),
        status_code=200,
        text=example_response('outbound_shipment/get_fulfillment_order.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    fulfillment_outbound_shipment = mws_client.fulfillment_outbound_shipment
    response = fulfillment_outbound_shipment.get_fulfillment_order(
        SellerFulfillmentOrderId='extern_id_1154539615776'
    )
    shipment_1, shipment_2 = response.FulfillmentShipment.member
    assert shipment_1.FulfillmentShipmentStatus == 'PENDING'
    assert len(shipment_1.FulfillmentShipmentItem.member) == 1
    assert shipment_1.FulfillmentShipmentItem.member[0].SellerSKU == 'SKU100'
    assert shipment_1.FulfillmentShipmentItem.member[0].Quantity == 2

    assert shipment_2.FulfillmentShipmentStatus == 'SHIPPED'
    assert len(shipment_2.FulfillmentShipmentItem.member) == 1
    assert shipment_2.FulfillmentShipmentItem.member[0].SellerSKU == 'SKU101'
    assert shipment_2.FulfillmentShipmentItem.member[0].Quantity == 1
    shipment_2_package, = shipment_2.FulfillmentShipmentPackage.member
    assert shipment_2_package.TrackingNumber == '93ZZ00'
    assert shipment_2_package.CarrierCode == 'UPS'



class Orders(object):
    """
    Implements an order API client for Amazon MWS

    .. code-block:: python

        response = client.orders.list_orders(CreatedAfter=start_date)
        for order in response.Orders.getchildren():
            print(order.AmazonOrderId)

    for other attributes of the order, refer to the Amazon MWS documentation.

    To fetch the next page of orders:

    .. code-block:: python

        response = client.orders.list_orders(CreatedAfter=start_date)
        page2 = client.orders.list_orders_by_next_token(
            response.NextToken
        )
    """
    VERSION = '2013-09-01'
    URI = '/Orders/' + VERSION

    def __init__(self, client):
        self.client = client

    def list_orders(self, **kwargs):
        """
        Returns orders created or updated during a time frame that you specify.

        `Learn more <http://docs.developer.amazonservices.com/en_US/orders-2013-09-01/Orders_ListOrders.html>`__
        """     # noqa: E501
        if 'MarketplaceId.Id.1' not in kwargs:
            # Not a single marketplace id is specified.
            # fallback to the default marketplace
            kwargs['MarketplaceId.Id.1'] = self.client.marketplace.id

        return self.client.get(
            'ListOrders', self.URI, kwargs, self.VERSION
        )

    def list_orders_by_next_token(self, NextToken):
        """
        Returns the next page of orders using the NextToken parameter.

        `Learn more <http://docs.developer.amazonservices.com/en_US/orders-2013-09-01/Orders_ListOrdersByNextToken.html>`__

        The `NextToken` argument is not very pythonic, but keeps the
        package user side code consistent with Amazon's documentation
        and avoids a surprise for the user.
        """     # noqa: E501
        return self.client.get(
            'ListOrdersByNextToken', self.URI,
            {'NextToken': NextToken}, self.VERSION
        )

    def get_order(self, AmazonOrderId):
        """
        Returns orders based on the AmazonOrderId values that you specify.

        `Learn more <http://docs.developer.amazonservices.com/en_US/orders-2013-09-01/Orders_GetOrder.html>`__
        """     # noqa: E501
        return self.client.get(
            'GetOrder', self.URI,
            {'AmazonOrderId': AmazonOrderId}, self.VERSION
        )

    def list_order_items(self, AmazonOrderId):
        """
        Returns order items based on the AmazonOrderId that you specify.

        `Learn more <http://docs.developer.amazonservices.com/en_US/orders-2013-09-01/Orders_ListOrderItems.html>`__
        """     # noqa: E501
        return self.client.get(
            'ListOrderItems', self.URI,
            {'AmazonOrderId': AmazonOrderId}, self.VERSION
        )

    def list_order_items_by_next_token(self, NextToken):
        """
        Returns the next page of order items using the NextToken parameter.

        `Learn more <http://docs.developer.amazonservices.com/en_US/orders-2013-09-01/Orders_ListOrderItemsByNextToken.html>`__
        """     # noqa: E501
        return self.client.get(
            'ListOrderItemsByNextToken', self.URI,
            {'NextToken': NextToken}, self.VERSION
        )

    def get_service_status(self):
        """
        Returns the operational status of the Orders API section.

        `Learn more <http://docs.developer.amazonservices.com/en_US/orders-2013-09-01/MWS_GetServiceStatus.html>`__
        """     # noqa: E501
        return self.client.get(
            'GetServiceStatus', self.URI,
            {}, self.VERSION
        )



class Products(object):

    def __init__(self, client):
        self.client = client

    def list_matching_products(self, query, query_context_id):
        """
        Returns a list of products and their attributes, based on a
        search query.

        Learn more: http://docs.developer.amazonservices.com/en_US/products/Products_ListMatchingProducts.html
        """  # noqa: E501
        pass



from decimal import Decimal


def test_list_orders(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Orders/2013-09-01',
        status_code=200,
        text=example_response('orders/list_orders.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.orders.list_orders()
    assert len(response.Orders.getchildren()) == 3
    assert response.Orders.Order[0].AmazonOrderId == '111-1234567-0000001'
    assert response.NextToken


def test_list_orders_by_next_token(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Orders/2013-09-01',
        status_code=200,
        text=example_response('orders/list_orders_by_next_token.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.orders.list_orders_by_next_token('TOKEN')
    assert len(response.Orders.getchildren()) == 2
    assert response.Orders.Order[0].AmazonOrderId == '111-1234567-0000002'
    assert response.NextToken


def test_list_order_items(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Orders/2013-09-01',
        status_code=200,
        text=example_response('orders/list_order_items.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.orders.list_order_items('111-1234567-0000001')
    assert len(response.OrderItems.getchildren()) == 1
    assert response.AmazonOrderId == '111-1234567-0000001'
    item1 = response.OrderItems.OrderItem[0]
    assert item1.ASIN == 'B0X1X2X3X4X5'
    assert Decimal(item1.ItemPrice.Amount.text) == Decimal('139.00')


def test_get_service_status(mws_client, mock_adapter, example_response):
    mock_adapter.register_uri(
        'GET',
        mws_client.marketplace.endpoint + '/Orders/2013-09-01',
        status_code=200,
        text=example_response('orders/get_service_status.xml'),
        headers={'Content-Type': 'text/xml'}
    )
    response = mws_client.orders.get_service_status()
    assert response.Status == 'GREEN'



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=7.0',
    'lxml',
    'requests',

    # Drop these when py2 support is dropped
    'future',
    'six',
    'unicodecsv',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Fulfil.IO Inc.",
    author_email='help@fulfil.io',
    python_requires='>=2.7',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 2.7',
    ],
    description="Python client for Amazon Marketplace Web Services (MWS)",
    entry_points={
        'console_scripts': [
            'pymws=pymws.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='pymws',
    name='pymws',
    packages=find_packages(include=['pymws', 'pymws.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/fulfilio/pymws',
    version='0.1.5',
    zip_safe=False,
)
